@@ -4,12 +4,7 @@
 // Licensed under the MIT license <LICENSE-MIT or http://opensource.org/licenses/MIT>, at your option. This file may not
 // be copied, modified, or distributed except according to those terms.
 
-// TODO test_no_loggy?
-// TODO: proc macros
-// TODO: check debug vs. trace
-
-//! An opinionated library for developing and testing rust applications that use logging.
-
+#![doc = include_str!("../README.md")]
 #![deny(warnings)]
 #![deny(rust_2018_idioms)]
 #![deny(clippy::all)]
@@ -18,16 +13,18 @@
 #![deny(clippy::nursery)]
 #![deny(clippy::cargo)]
 
-pub use loggy_macros::loggy;
+pub use loggy_macros::scope;
 
-use crate as loggy;
 use lazy_static::lazy_static;
-use log::{Level, LevelFilter, Log, Metadata, Record};
-use parking_lot::{Mutex, Once};
+use log::{logger, set_logger, set_max_level, Level, LevelFilter, Log, Metadata, Record};
+use parking_lot::Mutex;
 use std::cell::Cell;
 use std::fmt::Write;
 use std::io::{stderr, Write as IoWrite};
+use std::marker::PhantomData;
 use std::panic::{catch_unwind, set_hook, take_hook, AssertUnwindSafe};
+use std::sync::atomic::AtomicBool;
+use std::sync::Arc;
 use unindent::unindent;
 
 /// Log a structured message.
@@ -105,24 +102,30 @@ macro_rules! log {
     };
 }
 
-/// Log a countable error message.
+/// Log an error message, which will trigger a panic at the end of the current [`scope`].
 ///
 /// This is identical to invoking `log!(log::Level::Error, ...)`.
 ///
-/// Note that error messages are special. By default they are converted to a `std::panic!`, unless running inside a
-/// `[#loggy]` test or inside [`count_errors`].
-#[macro_export]
-macro_rules! error { ( $( $arg:tt )* ) => { loggy::log!( log::Level::Error , $( $arg )* ) } }
-
-/// Log an error message when it is known we are not counting errors.
+/// # Panics
 ///
-/// This allows the compiler to know that any following code is unreachable (which isn't always the case for errors).
+/// If invoked outside any scope, this will immediately panic with a generic message.
+#[macro_export]
+macro_rules! error { ( $( $arg:tt )* ) => {
+    loggy::log!( log::Level::Error , $( $arg )* );
+} }
+
+/// Generate a panic error message and immediately terminate the current thread.
+///
+/// # Panics
+///
+/// Always.
 #[macro_export]
 macro_rules! panic {
     ( $( $arg:tt )* ) => {
         {
+            loggy::force_panic();
             loggy::error!( $( $arg )* );
-            std::panic!("counting an uncountable error");
+            std::unreachable!();
         }
     }
 }
@@ -142,7 +145,11 @@ macro_rules! warn { ( $( $arg:tt )* ) => { loggy::log!( log::Level::Warn , $( $a
 #[macro_export]
 macro_rules! note {
     ( $is_error:expr, $( $arg:tt )* ) => {
-        loggy::log!( if $is_error { log::Level::Error } else { log::Level::Warn } , $( $arg )* )
+        if $is_error {
+            loggy::error!( $( $arg )* );
+        } else {
+            loggy::warn!( $( $arg )* );
+        }
     };
 }
 
@@ -156,13 +163,14 @@ macro_rules! info { ( $( $arg:tt )* ) => { loggy::log!( log::Level::Info , $( $a
 ///
 /// This is identical to invoking `log!(log::Level::Debug, ...)`.
 ///
-/// Debug messages are special. The are always emitted in debug builds, regardless of the requested log level. They are
-/// not captured by tests, and instead are always sent to the standard error. The idea being that debug messages are
-/// used for, well, debugging.
+/// Debug messages are special. The are always emitted in debug builds, regardless of the requested
+/// log level. They are never captured by tests, and instead are always sent to the standard error.
+/// The idea being that debug messages are used for, well, debugging.
 #[macro_export]
 macro_rules! debug { ( $( $arg:tt )* ) => { loggy::log!( log::Level::Debug , $( $arg )* ) } }
 
-/// Log a debugging message..
+/// Log a debugging message that must be removed from the code once some debugging task is
+/// complete.
 ///
 /// This is identical to invoking `debug!(...)`. Renaming it `todox!` ensures all uses will be reported by `cargo
 /// todox`, to ensure their removal once their usefulness is past.
@@ -172,24 +180,20 @@ macro_rules! todox { ( $( $arg:tt )* ) => { loggy::debug!( $( $arg )* ) } }
 /// Log a tracing message.
 ///
 /// This is identical to invoking `log!(log::Level::Trace, ...)`.
-///
-/// This is originally meant for more-detailed-than-debugging messages, which doesn't really fit the model of "debug
-/// messages are used for actual debugging". It may be useful if you want debug-like (very detailed) messages that are
-/// not subject to the special handling of actual debug messages.
 #[macro_export]
 macro_rules! trace { ( $( $arg:tt )* ) => { loggy::log!( log::Level::Trace , $( $arg )* ) } }
 
 /// Provide program-wide control over whether some note(s) are considered to be an error.
 ///
-/// This is typically invoked as `pub mod foo { is_an_error!(true /* or false */); }`.
+/// This is typically invoked as `pub mod some_condition { is_an_error!(true /* or false */); }`.
 ///
 /// This will define a nested module which exposes the following:
 ///
 /// * `pub fn is_an_error() -> bool` returns whether the event is considered to be an error. This is initialized to the
 ///   value passed to the macro.
 ///
-///   This is intended to be used in `note!(foo::is_an_error(), ...)`, using the `note!` macro provided by `loggy`. It
-///   will behave as either an `error!` or a `warn!` depending on the value.
+///   This is intended to be used in [`note!`], which will behave as either an [`error!`] or a
+///   [`warn!`] depending on the value.
 ///
 /// * `pub fn set_is_an_error(new_value: bool) -> bool` sets whether the event is considered to be an error, returning
 ///   the old value for convenience.
@@ -199,10 +203,8 @@ macro_rules! trace { ( $( $arg:tt )* ) => { loggy::log!( log::Level::Trace , $(
 #[macro_export]
 macro_rules! is_an_error {
     ($default:expr) => {
-        lazy_static! {
-            static ref IS_AN_ERROR: std::sync::atomic::AtomicBool =
-                std::sync::atomic::AtomicBool::new($default);
-        }
+        static IS_AN_ERROR: std::sync::atomic::AtomicBool =
+            std::sync::atomic::AtomicBool::new($default);
 
         pub fn set_is_an_error(is_error: bool) -> bool {
             IS_AN_ERROR.swap(is_error, std::sync::atomic::Ordering::Relaxed)
@@ -214,6 +216,69 @@ macro_rules! is_an_error {
     };
 }
 
+/// A named scope for log messages and [`error`]s.
+#[derive(Clone, Copy)]
+struct NamedScope {
+    /// The scope name (to replace the module name in the messages).
+    name: &'static str,
+
+    /// The number of errors we've seen in the scope.
+    errors: usize,
+}
+
+thread_local! {
+    static NAMED_SCOPE: Cell<Option<NamedScope>> = Cell::new(None);
+}
+
+/// An RAII scope for log messages and [`error`]s.
+pub struct Scope<'a> {
+    /// The previous scope in effect before this one.
+    previous: Option<NamedScope>,
+
+    /// Ensure the scope name outlives the scope.
+    name_lifetime: PhantomData<&'a str>,
+}
+
+impl<'a> Scope<'a> {
+    /// Create a new logging scope.
+    #[must_use]
+    pub fn new(name: &'a str) -> Self {
+        let name_ptr: *const str = name;
+        let static_name_ref: &'static str = unsafe { &*name_ptr };
+        let next: NamedScope = NamedScope {
+            name: static_name_ref,
+            errors: 0,
+        };
+        let previous = NAMED_SCOPE.with(|named_scope| named_scope.replace(Some(next)));
+        Scope {
+            previous,
+            name_lifetime: PhantomData,
+        }
+    }
+
+    /// Execute some code with in a named scope.
+    pub fn with<T, Code: FnOnce() -> T>(name: &'a str, code: Code) -> T {
+        let _scope = Scope::new(name);
+        code()
+    }
+}
+
+impl<'a> Drop for Scope<'a> {
+    fn drop(&mut self) {
+        let current = NAMED_SCOPE
+            .with(|named_scope| named_scope.replace(self.previous))
+            .unwrap();
+        if current.errors > 0 {
+            std::panic!(
+                "{}: [ERROR] {}: failed with {} error(s)",
+                Loggy::global().prefix,
+                current.name,
+                current.errors
+            );
+        }
+    }
+}
+
 /// Control the behavior of the `loggy` logger.
 pub struct Loggy {
     /// A prefix appended to each message.
@@ -229,23 +294,22 @@ pub struct Loggy {
 }
 
 lazy_static! {
-    // BEGIN NOT TESTED
+    /// The format to use for the time in emitted log messages.
     static ref TIME_FORMAT: Vec<time::format_description::FormatItem<'static>> =
-        time::format_description::parse("[year]-[month]-[day] [hour]:[minute]:[second]").unwrap();
-    // END NOT TESTED
-    static ref TOTAL_THREADS: std::sync::atomic::AtomicUsize = std::sync::atomic::AtomicUsize::new(0);
-    static ref TOTAL_ERRORS: std::sync::atomic::AtomicUsize =
-        std::sync::atomic::AtomicUsize::new(0);
+        time::format_description::parse("[year]-[month]-[day] [hour]:[minute]:[second]").unwrap(); // NOT TESTED
 }
 
+static TOTAL_THREADS: std::sync::atomic::AtomicUsize = std::sync::atomic::AtomicUsize::new(0);
+
 thread_local!(
     static THREAD_ID: Cell<Option<usize>> = Cell::new(None);
-    static NAMED_SCOPE: Cell<Option<&'static str>> = Cell::new(None);
-    static THREAD_ERRORS: Cell<usize> = Cell::new(0);
-    static IS_COUNTING_ERRORS: Cell<bool> = Cell::new(false);
+    static FORCE_PANIC: Cell<bool> = Cell::new(false);
 );
 
+// BEGIN APPEARS NOT TESTED
 impl Log for Loggy {
+    // END APPEARS NOT TESTED
+
     fn enabled(&self, metadata: &Metadata<'_>) -> bool {
         metadata.level() == Level::Error
             || metadata.level() == Level::Debug
@@ -253,14 +317,7 @@ impl Log for Loggy {
     }
 
     fn log(&self, record: &Record<'_>) {
-        if record.level() == Level::Error {
-            TOTAL_ERRORS.fetch_add(1, std::sync::atomic::Ordering::Relaxed);
-            THREAD_ERRORS.with(|thread_errors_cell| {
-                thread_errors_cell.set(thread_errors_cell.get() + 1);
-            });
-        }
-
-        if record.level() == Level::Error || self.enabled(record.metadata()) {
+        if self.enabled(record.metadata()) {
             emit_message(record.level(), self.format_message(record).as_ref());
         }
     }
@@ -272,41 +329,16 @@ impl Log for Loggy {
     // END NOT TESTED
 }
 
-/// Execute some code in a named scope.
-///
-/// This scope `name` will be used to prefix log messages generated by the `code`, instead of the default (module name)
-/// scope. An empty name will eliminate the prefix altogether.
-///
-/// Note this only affects error messages generated in the current thread.
-pub fn in_named_scope<Code: FnOnce()>(name: &'static str, code: Code) {
-    let old_scope = NAMED_SCOPE.with(|named_scope| named_scope.replace(Some(name)));
-    code();
-    NAMED_SCOPE.with(|named_scope| named_scope.set(old_scope));
-}
-
-/// Execute some code, while counting the errors.
-///
-/// Returns the number of errors reported by the `code`. These errors are also added to any containing scope. By
-/// default, unless running in a test, error (formatted) messages are passed to `std::panic!`. This is disabled when
-/// counting errors, allowing the code to emit multiple such messages. It is the caller's responsibility to examine the
-/// number of errors and do a final `panic!`, or otherwise handle the situation.
-///
-/// Note this only counts error messages generated in the current thread.
-#[must_use]
-pub fn count_errors<Code: FnOnce()>(code: Code) -> usize {
-    let was_counting_before =
-        IS_COUNTING_ERRORS.with(|is_counting_errors| is_counting_errors.replace(true)); // APPEARS NOT TESTED
-    let errors_before = THREAD_ERRORS.with(std::cell::Cell::get);
-
-    code();
-
-    IS_COUNTING_ERRORS.with(|is_counting_errors| is_counting_errors.set(was_counting_before));
-    let errors_after = THREAD_ERRORS.with(std::cell::Cell::get);
-
-    errors_after - errors_before
-}
-
 impl Loggy {
+    fn global() -> &'static Self {
+        let logger_ptr: *const dyn Log = logger();
+        #[allow(clippy::cast_ptr_alignment)]
+        #[allow(clippy::ptr_as_ptr)]
+        unsafe {
+            &*(logger_ptr as *const Self)
+        }
+    }
+
     fn format_message(&self, record: &Record<'_>) -> String {
         let now = if self.show_time {
             // BEGIN NOT TESTED
@@ -362,18 +394,20 @@ impl Loggy {
         write!(&mut message, " [{}]", level).unwrap();
 
         if record.level() == Level::Debug {
+            // BEGIN MAYBE TESTED
             write!(
+                // END MAYBE TESTED
                 &mut message,
                 " {}:{}:",
-                record.file().unwrap(),
-                record.line().unwrap()
+                record.file().unwrap(), // MAYBE TESTED
+                record.line().unwrap()  // MAYBE TESTED
             )
             .unwrap();
         }
 
         let scope = NAMED_SCOPE.with(|named_scope| match named_scope.get() {
             None => record.module_path().unwrap(),
-            Some(scope) => scope,
+            Some(scope) => scope.name,
         });
         if !scope.is_empty() {
             write!(&mut message, " {}:", scope).unwrap();
@@ -381,92 +415,52 @@ impl Loggy {
     }
 }
 
-enum LogSink {
-    Stderr,
-    Buffer,
-}
-
 lazy_static! {
+    /// The buffer capturing the log messages for assertions.
     static ref LOG_BUFFER: Mutex<Cell<Option<String>>> = Mutex::new(Cell::new(None));
-}
 
-fn set_log_sink(log_sink: &LogSink) {
-    match log_sink {
-        LogSink::Stderr => {
-            LOG_BUFFER.lock().set(None);
-        }
-        LogSink::Buffer => {
-            LOG_BUFFER.lock().set(Some(String::new()));
-        }
-    };
-}
-
-/// Assert that the collected log messages are as expected.
-///
-/// The expected string is passed through `unindent` prior to the comparison, to enable proper indentation of the tests
-/// data in the code.
-///
-/// This clears the log buffer following the comparison.
-///
-/// This is meant to be used in tests using the `[#loggy]` macro. Tests using this macro expect the log buffer being
-/// clear at the end of the test, either by using this function or `ignore_log`.
-///
-/// # Panics
-///
-/// If the actual log is different from the expected log.
-pub fn assert_logged(expected_log: &str) {
-    let mut log_buffer = LOG_BUFFER.lock();
-    match log_buffer.get_mut() {
-        None => {
-            std::panic!("asserting log when logging to stderr"); // NOT TESTED
-        }
-        Some(actual_log) => {
-            let expected_log = fix_expected(expected_log);
-            if *actual_log != expected_log {
-                // BEGIN NOT TESTED
-                print!(
-                    "ACTUAL LOG:\n>>>\n{}<<<\nIS DIFFERENT FROM EXPECTED LOG:\n>>>\n{}<<<\n",
-                    actual_log, expected_log
-                );
-                assert_eq!("ACTUAL LOG", "EXPECTED LOG");
-            } // END NOT TESTED
-            actual_log.clear();
-        }
-    }
-}
-
-/// Ignore (clear) the content of the captured log.
-///
-/// This is meant to be used in tests using the `[#loggy]` macro, if they do not otherwise consume the captured log
-/// (e.g. by calling [`assert_logged`]).
-///
-/// # Panics
-///
-/// If no log is being collected.
-pub fn ignore_log() {
-    let mut log_buffer = LOG_BUFFER.lock();
-    match log_buffer.get_mut() {
-        None => std::panic!("clearing log when logging to stderr"), // NOT TESTED
-        Some(ref mut buffer) => buffer.clear(),
-    }
-}
-
-lazy_static! {
+    /// Whether to mirror captured log messages to stderr.
     static ref MIRROR_TO_STDERR: bool = std::env::var("LOGGY_MIRROR_TO_STDERR")
-        // BEGIN NOT TESTED
+        // BEGIN MAYBE TESTED
         .map(|var| !var.is_empty())
         .unwrap_or(false);
-        // END NOT TESTED
+    // END MAYBE TESTED
 }
 
+/// Whether we already setup loggy as the global logger.
+static DID_SET_LOGGER: AtomicBool = AtomicBool::new(false);
+
+/// Force the next error-level message to be emitted as a panic.
+#[doc(hidden)]
+pub fn force_panic() {
+    FORCE_PANIC.with(|force_panic| {
+        force_panic.set(true);
+    });
+}
+
+/// Actually emit (or capture) a log message.
 fn emit_message(level: Level, message: &str) {
     if level == Level::Debug {
-        eprint!("{}", message);
+        eprint!("{}", message); // MAYBE TESTED
         return;
     }
 
-    if level == Level::Error && !IS_COUNTING_ERRORS.with(std::cell::Cell::get) {
-        std::panic!("{}", message);
+    if level == Level::Error {
+        if FORCE_PANIC.with(|forced_panic| forced_panic.replace(false)) {
+            std::panic!("{}", message);
+        } else {
+            NAMED_SCOPE.with(|maybe_named_scope| {
+                if let Some(ref mut named_scope) = maybe_named_scope.get() {
+                    named_scope.errors += 1;
+                    maybe_named_scope.set(Some(*named_scope));
+                } else {
+                    std::panic!(
+                        "{}: error! called outside a named scope",
+                        Loggy::global().prefix,
+                    );
+                }
+            });
+        }
     }
 
     let mut log_buffer = LOG_BUFFER.lock();
@@ -474,54 +468,65 @@ fn emit_message(level: Level, message: &str) {
         None => eprint!("{}", message), // NOT TESTED
         Some(buffer) => {
             if *MIRROR_TO_STDERR {
-                eprint!("{}", message); // NOT TESTED
+                eprint!("{}", message); // MAYBE TESTED
             }
             buffer.push_str(message);
         }
     }
 }
 
-/// Return the total number of calls to `error!` in the whole program.
+/// RAII for capturing the log content.
+struct Capture {}
+
+impl Capture {
+    fn new() -> Self {
+        if !DID_SET_LOGGER.swap(true, std::sync::atomic::Ordering::Relaxed) {
+            set_logger(&Loggy {
+                prefix: "test",
+                show_time: false,
+                show_thread: false,
+            })
+            .unwrap();
+            set_max_level(LevelFilter::Trace);
+        }
+
+        LOG_BUFFER.lock().set(Some(String::new()));
+
+        TOTAL_THREADS.store(0, std::sync::atomic::Ordering::Relaxed);
+        THREAD_ID.with(|thread_id_cell| thread_id_cell.set(None));
+
+        Self {}
+    }
+}
+
+impl Drop for Capture {
+    fn drop(&mut self) {
+        LOG_BUFFER.lock().set(None);
+    }
+}
+
+/// Assert that the collected log messages are as expected.
 ///
-/// This is reset for each test using the `[#loggy]` macro.
-#[must_use]
-pub fn errors() -> usize {
-    TOTAL_ERRORS.load(std::sync::atomic::Ordering::Relaxed)
-}
-
-static LOGGER_ONCE: Once = Once::new();
-
-#[doc(hidden)]
-pub fn before_test() {
-    LOGGER_ONCE.call_once(|| {
-        log::set_logger(&Loggy {
-            prefix: "test",
-            show_time: false,
-            show_thread: false,
-        })
-        .unwrap();
-        log::set_max_level(LevelFilter::Trace);
-    });
-
-    TOTAL_THREADS.store(0, std::sync::atomic::Ordering::Relaxed);
-    TOTAL_ERRORS.store(0, std::sync::atomic::Ordering::Relaxed);
-    THREAD_ID.with(|thread_id_cell| thread_id_cell.set(None));
-    THREAD_ERRORS.with(|thread_errors_cell| thread_errors_cell.set(0));
-    IS_COUNTING_ERRORS.with(|is_counting_errors| is_counting_errors.set(true));
-
-    set_log_sink(&LogSink::Buffer);
-}
-
-#[doc(hidden)]
-pub fn after_test() {
-    assert_logged("");
-    set_log_sink(&LogSink::Stderr);
-    IS_COUNTING_ERRORS.with(|is_counting_errors| is_counting_errors.set(false));
-}
-
-/// Ensure that executing some code will panic with a specific error message (ignoring and clearing the log).
+/// The expected string is passed through `unindent` prior to the comparison, to enable proper indentation of the tests
+/// data in the code.
 ///
-/// TODO: This isn't really the best place for this, but it is necessary to capture tests that `panic!`.
+/// # Notes
+///
+/// The rust `log` facade mandates using a single global logger. Therefore, only one test can capture the log at any
+/// given time, using a a global `Mutex`. Therefore, nesting this inside itself, [`assert_panics`] or
+/// [`assert_logs_panics`] will deadlock.
+///
+/// # Panics
+///
+/// If the actual log is different from the expected log.
+pub fn assert_logs<Code: FnOnce() -> Result, Result>(expected_log: &str, code: Code) -> Result {
+    let _single_test = SINGLE_TEST.lock();
+    do_assert_logs_panics(Some(expected_log), None, code).unwrap()
+}
+
+/// Ensure that executing some code will panic with a specific error message (ignoring the log).
+///
+/// TODO: This crate isn't really the best place for this.
 ///
 /// Unlike `#[should_panic(expected = "...")]`, this:
 /// * Allows isolating a specific part of the test instead of decorating a whole test;
@@ -529,64 +534,104 @@ pub fn after_test() {
 /// * Allows the caller to dynamically generate the expected panic message;
 /// * Insists on the exact panic string rather than just a sub-string if it.
 ///
+/// # Notes
+///
+/// The rust `log` facade mandates using a single global logger. Therefore, only one test can capture the log at any
+/// given time, using a a global `Mutex`. Therefore, nesting this inside itself, [`assert_logs`] or
+/// [`assert_logs_panics`] will deadlock.
+///
 /// # Panics
 ///
 /// If the code does not panic, or panics with a different message than expected.
 pub fn assert_panics<Code: FnOnce() -> Result, Result>(expected_panic: &str, code: Code) {
-    do_assert_logged_panics(None, expected_panic, code);
+    let _single_test = SINGLE_TEST.lock();
+    do_assert_logs_panics(None, Some(expected_panic), code);
 }
 
-/// Same as [`assert_panics`] but also assert that the log contained the specified entries at the point of the panic.
-pub fn assert_logged_panics<Code: FnOnce() -> Result, Result>(
+/// Combine [`assert_logs`] and [`assert_panics`], that is, assert that the expected log is generated and then the
+/// expected panic is triggered.
+///
+/// # Notes
+///
+/// The rust `log` facade mandates using a single global logger. Therefore, only one test can capture the log at any
+/// given time, using a a global `Mutex`. Therefore, nesting this inside itself, [`assert_panics`] or
+/// [`assert_logs_panics`] will deadlock.
+///
+/// # Panics
+///
+/// If the code does generate the expected log, or does not panic, or panics with a different message than expected.
+pub fn assert_logs_panics<Code: FnOnce() -> Result, Result>(
     expected_log: &str,
     expected_panic: &str,
     code: Code,
 ) {
-    do_assert_logged_panics(Some(expected_log), expected_panic, code);
+    let _single_test = SINGLE_TEST.lock();
+    do_assert_logs_panics(Some(expected_log), Some(expected_panic), code);
 }
 
-fn do_assert_logged_panics<Code: FnOnce() -> Result, Result>(
+lazy_static! {
+    /// Ensure there is only a single test which is capturing log entries.
+    static ref SINGLE_TEST: Arc<Mutex<bool>> = Arc::new(Mutex::new(false));
+}
+
+fn do_assert_logs_panics<Code: FnOnce() -> Result, Result>(
     expected_log: Option<&str>,
-    expected_panic: &str,
+    expected_panic: Option<&str>,
     code: Code,
-) {
-    let prev_hook = take_hook();
-    set_hook(Box::new(|_| {}));
+) -> Option<Result> {
+    let _capture = Capture::new();
 
-    let was_counting_before =
-        IS_COUNTING_ERRORS.with(|is_counting_errors| is_counting_errors.replace(false)); // APPEARS NOT TESTED
-    let result = catch_unwind(AssertUnwindSafe(code));
-    IS_COUNTING_ERRORS.with(|is_counting_errors| is_counting_errors.set(was_counting_before));
-    set_hook(prev_hook);
+    if let Some(expected_panic) = expected_panic {
+        let prev_hook = take_hook();
+        set_hook(Box::new(|_| {}));
+        let result = catch_unwind(AssertUnwindSafe(code));
+        set_hook(prev_hook);
 
-    if let Some(expected_log) = expected_log {
-        assert_logged(expected_log);
-    } else {
-        ignore_log();
-    }
+        do_assert_logs(expected_log);
 
-    match result {
-        Ok(_) => std::panic!("test did not panic"), // NOT TESTED
+        match result {
+            Ok(_) => std::panic!("test did not panic"), // NOT TESTED
 
-        Err(error) => {
-            let actual_panic = if let Some(actual_panic) = error.downcast_ref::<String>() {
-                actual_panic.as_str()
-            // BEGIN NOT TESTED
-            } else if let Some(actual_panic) = error.downcast_ref::<&'static str>() {
-                actual_panic
-            // END NOT TESTED
-            } else {
-                "unknown panic" // NOT TESTED
-            };
-            let expected_panic = fix_expected(expected_panic);
-            assert_eq!(actual_panic, expected_panic);
+            Err(error) => {
+                let actual_panic = if let Some(actual_panic) = error.downcast_ref::<String>() {
+                    actual_panic.as_str()
+                // BEGIN NOT TESTED
+                } else if let Some(actual_panic) = error.downcast_ref::<&'static str>() {
+                    actual_panic
+                // END NOT TESTED
+                } else {
+                    "unknown panic" // NOT TESTED
+                };
+                let expected_panic = fix_expected(expected_panic);
+                assert_eq!(actual_panic, expected_panic);
+            }
         }
+        None
+    } else {
+        let result = code();
+        do_assert_logs(expected_log);
+        Some(result)
+    }
+}
+
+fn do_assert_logs(expected_log: Option<&str>) {
+    if let Some(expected_log) = expected_log {
+        let actual_log = LOG_BUFFER.lock().take().unwrap();
+        let expected_log = fix_expected(expected_log);
+        if actual_log != expected_log {
+            // BEGIN NOT TESTED
+            print!(
+                "ACTUAL LOG:\n>>>\n{}<<<\nIS DIFFERENT FROM EXPECTED LOG:\n>>>\n{}<<<\n",
+                actual_log, expected_log
+            );
+            assert_eq!("ACTUAL LOG", "EXPECTED LOG");
+        } // END NOT TESTED
     }
 }
 
 /// Check that some code writes the expected results.
 ///
-/// TODO: This isn't really the best place for this.
+/// TODO: This crate isn't really the best place for this.
 ///
 /// # Panics
 ///
@@ -599,18 +644,6 @@ pub fn assert_writes<Code: FnOnce(&mut dyn IoWrite)>(expected_string: &str, code
     assert_eq!(actual_string, expected_string);
 }
 
-/// Assert that some code does not log any errors by counting them.
-///
-/// # Panics
-///
-/// If the code logs even one error.
-pub fn assert_no_errors<Code: FnOnce()>(name: &str, code: Code) {
-    let errors_count = count_errors(code);
-    if errors_count > 0 {
-        panic!("{} {} error(s)", errors_count, name);
-    }
-}
-
 fn fix_expected(expected: &str) -> String {
     let needs_trailing_newline = expected.ends_with('\n') || expected.ends_with(' ');
     let expected = unindent(expected);
